"""Module for a randomly acting RL agent."""
from typing import Union

import gym
from gym.spaces import Box, Discrete, MultiDiscrete, MultiBinary
import numpy as np

from irl_benchmark.config import RL_CONFIG_DOMAINS
from irl_benchmark.rl.algorithms.base_algorithm import BaseRLAlgorithm


class RandomAgent(BaseRLAlgorithm):
    """Random agent. Picks actions uniformly random.
    """

    def __init__(self, env: gym.Env, config: dict = None):
        """

        Parameters
        ----------
        env: gym.Env
            A gym environment
        config: dict
            Configuration of hyperparameters.
            Can be empty dictionary, random agent doesn't have hyperparams.
        """
        super(RandomAgent, self).__init__(env, config)
        assert hasattr(self.env.action_space, 'sample')

    def train(self, no_episodes: int):
        """ Random agent does not need to be trained."""
        pass

    def pick_action(self, state: Union[int, float, np.ndarray]
                    ) -> Union[int, float, np.ndarray]:
        """ Pick an action given a state.

        Picks uniformly random from all possible actions, using the environments
        action_space.sample() method.

        Parameters
        ----------
        state: int
            An integer corresponding to a state of a DiscreteEnv.
            Not used in this agent.

        Returns
        -------
        Union[int, float, np.ndarray]
            An action
        """
        # if other spaces are needed, check if their sample method conforms with
        # returned type, change if necessary.
        assert isinstance(self.env.action_space,
                          (Box, Discrete, MultiDiscrete, MultiBinary))
        return self.env.action_space.sample()

    def policy(self, state: int) -> np.ndarray:
        raise NotImplementedError()


RL_CONFIG_DOMAINS[RandomAgent] = {}



"""A module containing base classes for RL algorithms"""

from abc import ABC, abstractmethod
from typing import Union

import gym
import numpy as np

from irl_benchmark.config import preprocess_config, RL_CONFIG_DOMAINS


class BaseRLAlgorithm(ABC):
    """Base class for Reinforcement Learning agents."""

    def __init__(self, env: gym.Env, config: Union[dict, None] = None):
        """

        Parameters
        ----------
        env: gym.Env
            The gym environment the agent will be trained on.
        config: dict
            A dictionary of algorithm-specific parameters.
        """
        self.env = env
        self.config = preprocess_config(self, RL_CONFIG_DOMAINS, config)

    @abstractmethod
    def train(self, no_episodes: int):
        """ Train the agent.

        Parameters
        ----------
        no_episodes: int
            Training will be run for this number of episodes.
        """
        raise NotImplementedError()

    @abstractmethod
    def pick_action(self, state: Union[int, float, np.ndarray]
                    ) -> Union[int, float, np.ndarray]:
        """ Pick an action given a state.

        Parameters
        ----------
        state: Union[int, float, np.ndarray]
            A state of the environment, compatible with env.observation_space.

        Returns
        -------
        Union[int, float, np.ndarray]
            An action, compatible with env.action_space

        """
        raise NotImplementedError()

    @abstractmethod
    def policy(self, state: Union[int, float, np.ndarray]) -> np.ndarray:
        """Return the probabilities of picking actions given a state.

        NOTE: This is currently only defined for discrete action spaces.

        Parameters
        ----------
        state: Union[int, float, np.ndarray]
            A state of the environment, compatible with env.observation_space.

        Returns
        -------
        np.ndarray
            A numpy ndarray containing probabilities of each action.
            The shape of the array corresponds to the dimension of env.action_space.

        """
        raise NotImplementedError()



"""Util functions related to IRL."""

from typing import Dict, List
import numpy as np

from irl_benchmark.irl.feature.feature_wrapper import FeatureWrapper
from irl_benchmark.utils.wrapper import unwrap_env, is_unwrappable_to


def feature_count(env, trajs: List[Dict[str, list]],
                  gamma: float) -> np.ndarray:
    """Return empirical discounted feature counts of input trajectories.

    Parameters
    ----------
    env: gym.Env
        A gym environment, wrapped in a feature wrapper
    trajs: List[Dict[str, list]]
         A list of trajectories.
        Each trajectory is a dictionary with keys
        ['states', 'actions', 'rewards', 'true_rewards', 'features'].
        The values of each dictionary are lists.
        See :func:`irl_benchmark.irl.collect.collect_trajs`.
    gamma: float
        The discount factor. Must be in range [0., 1.].

    Returns
    -------
    np.ndarray
        A numpy array containing discounted feature counts. The shape
        is the same as the trajectories' feature shapes. One scalar
        feature count per feature.
    """
    assert is_unwrappable_to(env, FeatureWrapper)

    # Initialize feature count sum to zeros of correct shape:
    feature_count_sum = np.zeros(
        unwrap_env(env, FeatureWrapper).feature_shape())

    for traj in trajs:
        assert traj['features']  # empty lists are False in python

        # gammas is a vector containing [gamma^0, gamma^1, gamma^2, ... gamma^l]
        # where l is length of the trajectory:
        gammas = gamma**np.arange(len(traj['features']))
        traj_feature_count = np.sum(
            gammas.reshape(-1, 1) * np.array(traj['features']), axis=0)
        # add trajectory's feature count:
        feature_count_sum += traj_feature_count
    # divide feature_count_sum by number of trajectories to normalize:
    result = feature_count_sum / len(trajs)
    return result



"""Module for the abstract base class of all IRL algorithms."""
from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Tuple, Union

import gym
import numpy as np

from irl_benchmark.config import preprocess_config, IRL_CONFIG_DOMAINS
from irl_benchmark.irl.reward.reward_function import BaseRewardFunction
from irl_benchmark.irl.reward.reward_wrapper import RewardWrapper
from irl_benchmark.metrics.base_metric import BaseMetric
from irl_benchmark.rl.algorithms.base_algorithm import BaseRLAlgorithm
from irl_benchmark.utils.wrapper import is_unwrappable_to
import irl_benchmark.utils.irl as irl_utils


class BaseIRLAlgorithm(ABC):
    """The abstract base class for all IRL algorithms."""

    def __init__(self,
                 env: gym.Env,
                 expert_trajs: List[Dict[str, list]],
                 rl_alg_factory: Callable[[gym.Env], BaseRLAlgorithm],
                 metrics: List[BaseMetric] = None,
                 config: Union[dict, None] = None):
        """

        Parameters
        ----------
        env: gym.Env
            The gym environment to be trained on.
            Needs to be wrapped in a RewardWrapper to not leak the true reward function.
        expert_trajs: List[dict]
            A list of trajectories.
            Each trajectory is a dictionary with keys
            ['states', 'actions', 'rewards', 'true_rewards', 'features'].
            The values of each dictionary are lists.
            See :func:`irl_benchmark.irl.collect.collect_trajs`.
        rl_alg_factory: Callable[[gym.Env], BaseRLAlgorithm]
            A function which returns a new RL algorithm when called.
        config: dict
            A dictionary containing algorithm-specific parameters.
        """

        assert is_unwrappable_to(env, RewardWrapper)
        self.env = env
        self.expert_trajs = expert_trajs
        self.rl_alg_factory = rl_alg_factory
        if metrics is None:
            metrics = []
        self.metrics = metrics
        self.metric_results = [[]] * len(metrics)
        self.config = preprocess_config(self, IRL_CONFIG_DOMAINS, config)

    @abstractmethod
    def train(self, no_irl_iterations: int,
              no_rl_episodes_per_irl_iteration: int,
              no_irl_episodes_per_irl_iteration: int
              ) -> Tuple[BaseRewardFunction, BaseRLAlgorithm]:
        """Train the IRL algorithm.

        Parameters
        ----------
        no_irl_iterations: int
            The number of iteration the algorithm should be run.
        no_rl_episodes_per_irl_iteration: int
            The number of episodes the RL algorithm is allowed to run in
            each iteration of the IRL algorithm.
        no_irl_episodes_per_irl_iteration: int
            The number of episodes permitted to be run in each iteration
            to update the current reward estimate (e.g. to estimate state frequencies
            of the currently optimal policy).

        Returns
        -------
        Tuple[BaseRewardFunction, BaseRLAlgorithm]
            The estimated reward function and a RL agent trained for this estimate.
        """
        raise NotImplementedError()

    def evaluate_metrics(self, evaluation_input: dict):
        for metric in self.metrics:
            result = metric.evaluate(evaluation_input)
            self.metric_results.append(result)
            print(type(metric).__name__ + ': \t' + str(result))

    def feature_count(self, trajs: List[Dict[str, list]],
                      gamma: float) -> np.ndarray:
        """Return empirical discounted feature counts of input trajectories.

        Parameters
        ----------
        trajs: List[Dict[str, list]]
             A list of trajectories.
            Each trajectory is a dictionary with keys
            ['states', 'actions', 'rewards', 'true_rewards', 'features'].
            The values of each dictionary are lists.
            See :func:`irl_benchmark.irl.collect.collect_trajs`.
        gamma: float
            The discount factor. Must be in range [0., 1.].

        Returns
        -------
        np.ndarray
            A numpy array containing discounted feature counts. The shape
            is the same as the trajectories' feature shapes. One scalar
            feature count per feature.
        """
        # This was moved to utils:
        return irl_utils.feature_count(self.env, trajs, gamma)



"""Module for 2-norm feature count loss."""

import numpy as np

from irl_benchmark.metrics.base_metric import BaseMetric
from irl_benchmark.utils.irl import feature_count


class FeatureCount2Loss(BaseMetric):
    """2-norm feature count loss."""

    def __init__(self, metric_input: dict):
        assert 'env' in metric_input.keys()
        assert 'expert_trajs' in metric_input.keys()
        super(FeatureCount2Loss, self).__init__(metric_input)
        expert_trajs = metric_input['expert_trajs']
        self.env = metric_input['env']
        self.expert_feature_count = feature_count(
            self.env, expert_trajs, gamma=1.0)

    def evaluate(self, evaluation_input: dict = None) -> float:
        """Evaluate the metric given some input and return result.

        Parameters
        ----------
        evaluation_input: dict

        Returns
        -------
        float
            The metric's output
        """
        assert 'irl_agent' in evaluation_input.keys()
        irl_trajs = self.generate_traj_if_not_exists(evaluation_input)
        irl_feature_count = feature_count(self.env, irl_trajs, gamma=1.0)
        diff = self.expert_feature_count - irl_feature_count
        return np.linalg.norm(diff, ord=2)



"""Module for apprenticeship learning IRL."""
from typing import Callable, Dict, List, Tuple

import cvxpy as cvx
import gym
import numpy as np

from irl_benchmark.config import IRL_CONFIG_DOMAINS
from irl_benchmark.irl.algorithms.base_algorithm import BaseIRLAlgorithm
from irl_benchmark.irl.collect import collect_trajs
from irl_benchmark.irl.reward.reward_function import BaseRewardFunction, FeatureBasedRewardFunction
from irl_benchmark.irl.reward.reward_wrapper import RewardWrapper
from irl_benchmark.metrics.base_metric import BaseMetric
from irl_benchmark.rl.algorithms.base_algorithm import BaseRLAlgorithm
from irl_benchmark.rl.algorithms.random_agent import RandomAgent


class ApprIRL(BaseIRLAlgorithm):
    """
    Apprenticeship learning IRL (Abbeel & Ng, 2004).

    Assumes reward linear in features.
    Two variants are implemented: SVM-based and projection-based.
    """

    def __init__(self, env: gym.Env, expert_trajs: List[Dict[str, list]],
                 rl_alg_factory: Callable[[gym.Env], BaseRLAlgorithm],
                 metrics: List[BaseMetric], config: dict):
        """

        Parameters
        ----------
        env: gym.Env
            The gym environment to be trained on.
            Needs to be wrapped in a RewardWrapper to prevent leaking the true reward function.
        expert_trajs: List[dict]
            A list of trajectories.
            Each trajectory is a dictionary with keys
            ['states', 'actions', 'rewards', 'true_rewards', 'features'].
            The values of each dictionary are lists.
            See :func:`irl_benchmark.irl.collect.collect_trajs`.
        rl_alg_factory: Callable[[gym.Env], BaseRLAlgorithm]
            A function which returns a new RL algorithm when called.
        config: dict
            A dictionary containing hyper-parameters for the algorithm.
            The fields are:
            * 'gamma': discount factor between 0. and 1.
            * 'epsilon': small positive value, stopping criterion.
            * 'mode': which variant of the algorithm to use, either 'svm' or 'projection'.
        """
        super(ApprIRL, self).__init__(env, expert_trajs, rl_alg_factory,
                                      metrics, config)

        # calculate the feature counts of expert trajectories:
        self.expert_feature_count = self.feature_count(self.expert_trajs,
                                                       self.config['gamma'])

        # create list of feature counts:
        self.feature_counts = [self.expert_feature_count]
        # for SVM mode: create list of labels:
        self.labels = [1.]

        self.distances = []

    def train(self, no_irl_iterations: int,
              no_rl_episodes_per_irl_iteration: int,
              no_irl_episodes_per_irl_iteration: int
              ) -> Tuple[BaseRewardFunction, BaseRLAlgorithm]:
        """Train the apprenticeship learning IRL algorithm.

        Parameters
        ----------
        no_irl_iterations: int
            The number of iteration the algorithm should be run.
        no_rl_episodes_per_irl_iteration: int
            The number of episodes the RL algorithm is allowed to run in
            each iteration of the IRL algorithm.
        no_irl_episodes_per_irl_iteration: int
            The number of episodes permitted to be run in each iteration
            to update the current reward estimate (e.g. to estimate state frequencies
            of the currently optimal policy).

        Returns
        -------
        Tuple[BaseRewardFunction, BaseRLAlgorithm]
            The estimated reward function and a RL agent trained for this estimate.
        """

        # Initialize training with a random agent.
        agent = RandomAgent(self.env)

        irl_iteration_counter = 0
        while irl_iteration_counter < no_irl_iterations:
            irl_iteration_counter += 1

            if self.config['verbose']:
                print('IRL ITERATION ' + str(irl_iteration_counter))

            # Estimate feature count of current agent.
            trajs = collect_trajs(
                self.env,
                agent,
                no_trajectories=no_irl_episodes_per_irl_iteration)
            current_feature_count = self.feature_count(
                trajs, gamma=self.config['gamma'])

            # add new feature count to list of feature counts
            self.feature_counts.append(current_feature_count)
            # for SVM mode:
            self.labels.append(-1.)

            # convert to numpy array:
            feature_counts = np.array(self.feature_counts)
            labels = np.array(self.labels)

            # update reward coefficients based on mode specified in config:
            if self.config['mode'] == 'projection':
                # projection mode:
                if irl_iteration_counter == 1:
                    # initialize feature_count_bar in first iteration
                    # set to first non-expert feature count:
                    feature_count_bar = feature_counts[1]
                else:
                    # not first iteration.
                    # calculate line through last feature_count_bar and
                    # last non-expert feature count:
                    line = feature_counts[-1] - feature_count_bar
                    # new feature_count_bar is orthogonal projection of
                    # expert's feature count onto the line:
                    feature_count_bar += np.dot(
                        line, feature_counts[0] - feature_count_bar) / np.dot(
                            line, line) * line
                reward_coefficients = feature_counts[0] - feature_count_bar
                # compute distance as L2 norm of reward coefficients (t^(i) in paper):
                distance = np.linalg.norm(reward_coefficients, ord=2)

            elif self.config['mode'] == 'svm':
                # svm mode:
                # create quadratic programming problem definition:
                weights = cvx.Variable(feature_counts.shape[1])
                bias = cvx.Variable()
                objective = cvx.Minimize(cvx.norm(weights, 2))
                constraints = [
                    cvx.multiply(labels,
                                 (feature_counts * weights + bias)) >= 1
                ]
                problem = cvx.Problem(objective, constraints)
                # solve quadratic program:
                problem.solve()

                if weights.value is None:
                    # TODO: we need to handle empty solution better.
                    raise RuntimeError(
                        'Empty solution set for linearly separable SVM.')

                if self.config['verbose']:
                    # print support vectors
                    # (which last iterations where relevant for current result?)
                    svm_classifications = feature_counts.dot(
                        weights.value) + bias.value
                    support_vectors = np.where(
                        np.isclose(np.abs(svm_classifications), 1))[0]
                    print('The support vectors are from iterations number ' +
                          str(support_vectors))

                reward_coefficients = weights.value
                distance = 2 / problem.value

            else:
                raise NotImplementedError()

            if self.config['verbose']:
                print('Distance: ' + str(distance))

            self.distances.append(distance)

            # create new reward function with current coefficient estimate
            reward_function = FeatureBasedRewardFunction(
                self.env, reward_coefficients)
            # update reward function
            assert isinstance(self.env, RewardWrapper)
            self.env.update_reward_function(reward_function)

            # check stopping criterion:
            if distance <= self.config['epsilon']:
                if self.config['verbose']:
                    print("Feature counts matched within " +
                          str(self.config['epsilon']) + ".")
                break

            # create new RL-agent
            agent = self.rl_alg_factory(self.env)
            # train agent (with new reward function)
            agent.train(no_rl_episodes_per_irl_iteration)

            evaluation_input = {
                'irl_agent': agent,
                'irl_reward': reward_function
            }
            self.evaluate_metrics(evaluation_input)

        return reward_function, agent


IRL_CONFIG_DOMAINS[ApprIRL] = {
    'gamma': {
        'type': float,
        'min': 0.0,
        'max': 1.0,
        'default': 0.9,
    },
    'mode': {
        'type': 'categorical',
        'values': ['svm', 'projection'],
        'default': 'svm',
    },
    'epsilon': {
        'type': float,
        'min': 0.0,
        'max': float('inf'),
        'default': 1e-6,
    },
    'verbose': {
        'type': bool,
        'default': True
    }
}



"""This module contains configurations for IRL and RL algorithms."""

from copy import copy

# A dictionary containing allowed and default values for
# the config of each IRL algorithm.
# To be extended in each specific IRL algorithm implementation:
IRL_CONFIG_DOMAINS = {}

# A dictionary containing allowed and default values for
# the config of each RL algorithm.
# To be extended in each specific RL algorithm implementation:
RL_CONFIG_DOMAINS = {}


def preprocess_config(config_for: object, domains: dict, config: dict) -> dict:
    """ Pre-processes a config dictionary.

    This is based on the values specified for each IRL algorithm in
    IRL_CONFIG_DOMAINS. The following steps are performed:

    * If values in config are missing, add their default values.
    * If values are illegal (e.g. too high), raise an error.
    * If unknown fields are specified, raise an error.

    Manipulates a copy of the passed config and returns it.

    Parameters
    ----------
    config: dict
        The unprocessed config dictionary.

    Returns
    -------
    dict
        The processed config dictionary.

    """
    # replace config by empty dictionary if None:
    if config is None:
        config = {}
    else:
        config = copy(config)

    # get config domain for the correct subclass calling this:
    config_domain: dict = domains[type(config_for)]
    for key in config_domain.keys():
        if key in config.keys():
            # for numerical fields:
            if config[key] is None and 'optional' in config_domain[key].keys() \
                    and config_domain[key]['optional']:
                # encountered optional field with value None,
                # no checks necessary
                continue
            elif config_domain[key]['type'] in [float, int]:
                # check if right type:
                assert isinstance(
                    config[key], config_domain[key]
                    ['type']), "Wrong config value type for key " + str(key)
                # check if value is high enough:
                assert config[key] >= config_domain[key][
                    'min'], "Config value too low for key " + str(key)
                # check if value is low enough:
                assert config[key] <= config_domain[key][
                    'max'], "Config value too high for key " + str(key)
            # for categorical fields:
            elif config_domain[key]['type'] == 'categorical':
                # check if value is allowed:
                assert config[key] in config_domain[key][
                    'values'], "Illegal config value : " + config[key]
            elif config_domain[key]['type'] is bool:
                assert isinstance(config[key], bool)
            else:
                # encountered type for which no implementation has been written
                # extend code above to fix.
                raise NotImplementedError(
                    "No implementation for config value type: " +
                    str(config_domain[key]['type']))
        else:
            # key not specified in given config, use default value:
            config[key] = config_domain[key]['default']
    # check if config only contains legal fields:
    for key in config.keys():
        if key not in config_domain.keys():
            raise ValueError("Unknown config field: " + str(key))

    # return the pre-processed config:
    return config



"""Utils related to reinforcement learning."""
from typing import Dict, List

import numpy as np


def true_reward_per_traj(trajs: List[Dict[str, list]]) -> float:
    """Return (undiscounted) average sum of true rewards per trajectory.

    Parameters
    ----------
    trajs: List[Dict[str, list]])
        A list of trajectories.
        Each trajectory is a dictionary with keys
        ['states', 'actions', 'rewards', 'true_rewards', 'features'].
        The values of each dictionary are lists.
        See :func:`irl_benchmark.irl.collect.collect_trajs`.

    Returns
    -------
    float
        The undiscounted average sum of true rewards per trajectory.

    """
    true_reward_sum = 0
    for traj in trajs:
        true_reward_sum += np.sum(traj['true_rewards'])
    return true_reward_sum / len(trajs)



"""Module for maximum entropy inverse reinforcement learning."""

from typing import Callable, Dict, List

import gym
from gym.envs.toy_text.discrete import DiscreteEnv
import numpy as np

from irl_benchmark.config import IRL_CONFIG_DOMAINS
from irl_benchmark.irl.algorithms.base_algorithm import BaseIRLAlgorithm
from irl_benchmark.irl.feature.feature_wrapper import FeatureWrapper
from irl_benchmark.irl.reward.reward_function import FeatureBasedRewardFunction
from irl_benchmark.metrics.base_metric import BaseMetric
from irl_benchmark.rl.algorithms.base_algorithm import BaseRLAlgorithm
from irl_benchmark.utils.wrapper import get_transition_matrix, is_unwrappable_to, unwrap_env


class MaxEntIRL(BaseIRLAlgorithm):
    """Maximum Entropy IRL (Ziebart et al., 2008).

    Not to be confused with Maximum Entropy Deep IRL (Wulfmeier et al., 2016)
    or Maximum Causal Entropy IRL (Ziebart et al., 2010).
    """

    def __init__(self, env: gym.Env, expert_trajs: List[Dict[str, list]],
                 rl_alg_factory: Callable[[gym.Env], BaseRLAlgorithm],
                 metrics: List[BaseMetric], config: dict):
        """See :class:`irl_benchmark.irl.algorithms.base_algorithm.BaseIRLAlgorithm`."""

        assert is_unwrappable_to(env, DiscreteEnv)
        assert is_unwrappable_to(env, FeatureWrapper)

        super(MaxEntIRL, self).__init__(env, expert_trajs, rl_alg_factory,
                                        metrics, config)
        # get transition matrix (with absorbing state)
        self.transition_matrix = get_transition_matrix(self.env)
        self.n_states, self.n_actions, _ = self.transition_matrix.shape

        # get map of features for all states:
        feature_wrapper = unwrap_env(env, FeatureWrapper)
        self.feat_map = feature_wrapper.feature_array()

    def expected_svf(self, policy: np.ndarray) -> np.ndarray:
        """Calculate the expected state visitation frequency for the trajectories
        under the given policy. Returns vector of state visitation frequencies.
        Uses self.transition_matrix.

        Parameters
        ----------
        policy: np.ndarray
            The policy for which to calculate the expected SVF.

        Returns
        -------
        np.ndarray
            Expected state visitation frequencies as a numpy array of shape (n_states,).
        """
        # get the length of longest trajectory:
        longest_traj_len = 1  # init
        for traj in self.expert_trajs:
            longest_traj_len = max(longest_traj_len, len(traj['states']))

        # svf[state, time] is the frequency of visiting a state at some point of time
        svf = np.zeros((self.n_states, longest_traj_len))

        for traj in self.expert_trajs:
            svf[traj['states'][0], 0] += 1
        svf[:, 0] = svf[:, 0] / len(self.expert_trajs)

        for time in range(1, longest_traj_len):
            for state in range(self.n_states):
                total = 0
                for previous_state in range(self.n_states):
                    for action in range(self.n_actions):
                        total += svf[
                            previous_state, time - 1] * self.transition_matrix[
                                previous_state, action, state] * policy[
                                    previous_state, action]
                svf[state, time] = total
        # sum over all time steps and return SVF for each state:
        return np.sum(svf, axis=1)

    def train(self, no_irl_iterations: int,
              no_rl_episodes_per_irl_iteration: int,
              no_irl_episodes_per_irl_iteration: int):
        """Train algorithm. See abstract base class for parameter types."""

        # calculate feature expectations
        expert_feature_count = self.feature_count(self.expert_trajs, gamma=1.0)

        # start with an agent
        agent = self.rl_alg_factory(self.env)

        reward_function = FeatureBasedRewardFunction(self.env, 'random')
        self.env.update_reward_function(reward_function)
        theta = reward_function.parameters

        irl_iteration_counter = 0
        while irl_iteration_counter < no_irl_iterations:
            irl_iteration_counter += 1

            if self.config['verbose']:
                print('IRL ITERATION ' + str(irl_iteration_counter))
            # compute policy
            agent.train(no_rl_episodes_per_irl_iteration)

            policy = agent.policy_array()

            # compute state visitation frequencies, discard absorbing state
            svf = self.expected_svf(policy)[:-1]

            # compute gradients
            grad = (expert_feature_count - self.feat_map.T.dot(svf))

            # update params
            theta += self.config['lr'] * grad

            reward_function.update_parameters(theta)

            evaluation_input = {
                'irl_agent': agent,
                'irl_reward': reward_function
            }
            self.evaluate_metrics(evaluation_input)

        return theta


IRL_CONFIG_DOMAINS[MaxEntIRL] = {
    'verbose': {
        'type': bool,
        'default': True
    },
    'lr': {
        'type': float,
        'default': 0.02,
        'min': 0.000001,
        'max': 50
    }
}



"""Module for feature wrappers providing features for different environments."""

from abc import abstractmethod
import functools
from typing import Union, Tuple

import gym
import numpy as np

from irl_benchmark.utils.general import to_one_hot


class FeatureWrapper(gym.Wrapper):
    """Wrapper that adds features to the info dictionary in the step function.

    Generally, each environment needs its own feature wrapper."""

    def __init__(self, env: gym.Env):
        """

        Parameters
        ----------
        env: gym.Env
            The gym environment to be wrapped.
        """
        super(FeatureWrapper, self).__init__(env)
        self.current_state = None

    def reset(self, **kwargs):  # pylint: disable=E0202, R0801
        """ Reset environment and return initial state. No changes to base class reset function."""
        self.current_state = self.env.reset()
        return self.current_state

    def step(self, action: Union[np.ndarray, int, float]
             ) -> Tuple[Union[np.ndarray, float, int], float, bool, dict]:
        """

        Parameters
        ----------
        action: Union[np.ndarray, int, float]

        Returns
        -------
        Tuple[Union[np.ndarray, float, int], float, bool, dict]
            Tuple with values for (state, reward, done, info).
            Normal return values of any gym step function.
            A field 'features' is added to the returned info dict.
        """
        # pylint: disable=E0202

        # execute action:
        next_state, reward, terminated, info = self.env.step(action)

        info['features'] = self.features(self.current_state, action,
                                         next_state)

        # remember which state we are in:
        self.current_state = next_state

        return next_state, reward, terminated, info

    @abstractmethod
    def features(self, current_state: Union[np.ndarray, int, float],
                 action: Union[np.ndarray, int, float],
                 next_state: Union[np.ndarray, int, float]) -> np.ndarray:
        """Return features to be saved in step method's info dictionary."""
        raise NotImplementedError()

    @abstractmethod
    def feature_shape(self) -> tuple:
        """Get shape of features."""
        raise NotImplementedError()

    @abstractmethod
    def feature_range(self) -> np.ndarray:
        """Get minimum and maximum values of all k features.

        Returns
        -------
        np.ndarray
            The minimum and maximum values in an array of shape (2, k).
            First row corresponds to minimum values and second row to maximum values.
        """
        raise NotImplementedError()

    @abstractmethod
    def feature_array(self) -> np.ndarray:
        """ Get features for the entire domain as an array.
        Has to be overwritten in each feature wrapper.

        Returns
        -------
        np.ndarray
            The features for the entire domain as an array.
            Shape: (domain_size, feature_size).
        """
        raise NotImplementedError()


class FrozenLakeFeatureWrapper(FeatureWrapper):
    """Feature wrapper that was ad hoc written for the FrozenLake env.

    Would also work to get one-hot features for any other discrete env
    such that feature-based algorithms can be used in a tabular setting.
    """

    def features(self, current_state, action, next_state):
        '''Return one-hot encoding of next_state.'''
        return to_one_hot(next_state, self.env.observation_space.n)

    def feature_shape(self):
        '''Return dimension of the one-hot vectors used as features.'''
        return (self.env.observation_space.n, )

    def feature_range(self):
        '''Get maximum and minimum values of all k features.

        Returns:
        `np.ndarray` of shape (2, k) w/ max in 1st and min in 2nd row.
        '''
        ranges = np.zeros((2, self.feature_shape()[0]))
        ranges[1, :] = 1.0
        return ranges

    def feature_array(self):
        """Returns feature array for FrozenLake. Each state in the domain
        corresponds to a one_hot vector. Features of all states together
        are the identity matrix."""
        return np.eye(self.env.observation_space.n)


# # # # # #
# MAKE FEATURE WRAPPERS ACCESSIBLE BELOW:
# # # # # #

_FEATURE_WRAPPERS = {}


def feature_wrappable_envs() -> set:
    """Return list of ids for all gym environments that can currently be
    wrapped with a feature wrapper."""
    return set(_FEATURE_WRAPPERS.keys())


def make(key: str) -> FeatureWrapper:
    """Return a feature wrapper around the gym environment specified with key.

    Parameters
    ----------
    key: str
        A gym environment's id (can be found as env.spec.id),
        for example 'FrozenLake-v0'.

    Returns
    -------
    FeatureWrapper
        An environment created as gym.make(key) wrapped in an
        adequate feature wrapper.
    """
    # the _FEATURE_WRAPPERS dict is filled below by registering environments with
    # @_register_feature_wrapper.
    return _FEATURE_WRAPPERS[key]()


def _register_feature_wrapper(key: str):
    """Unified way of registering feature wrappers for gym environments."""

    def decorator(decorated_function):
        @functools.wraps(decorated_function)
        def wrapper_factory():
            # return a new feature wrapper around a new gym environment:
            return decorated_function(gym.make(key))

        # add docstring
        wrapper_factory.__doc__ = "Creates feature wrapper for {}".format(key)
        # add to list of feature-wrappable environments
        _FEATURE_WRAPPERS[key] = wrapper_factory
        return wrapper_factory

    return decorator


@_register_feature_wrapper('FrozenLake-v0')
def frozen_lake(env: gym.Env):
    """Register 'FrozenLake-v0' feature wrapper."""
    return FrozenLakeFeatureWrapper(env)


@_register_feature_wrapper('FrozenLake8x8-v0')
def frozen_lake_8_8(env: gym.Env):
    """Register 'FrozenLake-v2' feature wrapper."""
    # same feature wrapper as for 'FrozenLake-v0' can be used
    # as size of state space is automatically extracted
    return FrozenLakeFeatureWrapper(env)



"""Utils module containing wrapper specific helper functions."""

from copy import copy
from typing import Type, Union

import gym
from gym.envs.toy_text.discrete import DiscreteEnv
import numpy as np

from irl_benchmark.irl.reward.reward_wrapper import RewardWrapper

def unwrap_env(env: gym.Env,
               until_class: Union[None, gym.Env] = None) -> gym.Env:
    """Unwrap wrapped env until we get an instance that is a until_class.

    If until_class is None, env will be unwrapped until the lowest layer.
    """
    if until_class is None:
        while hasattr(env, 'env'):
            env = env.env
        return env

    while hasattr(env, 'env') and not isinstance(env, until_class):
        env = env.env

    if not isinstance(env, until_class):
        raise ValueError(
            "Unwrapping env did not yield an instance of class {}".format(
                until_class))
    return env


def is_unwrappable_to(env: gym.Env, to_wrapper: Type[gym.Wrapper]) -> bool:
    """Check if env can be unwrapped to to_wrapper.

    Parameters
    ----------
    env: gym.Env
        A gym environment (potentially wrapped).
    to_wrapper: Type[gym.Wrapper]
        A wrapper class extending gym.Wrapper.

    Returns
    -------
    bool
        True if env could be unwrapped to desired wrapper, False otherwise.
    """
    if isinstance(env, to_wrapper):
        return True
    while hasattr(env, 'env'):
        env = env.env
        if isinstance(env, to_wrapper):
            return True
    return False


def get_transition_matrix(env: gym.Env, with_absorbing_state: bool = True):
    """Get transition matrix from discrete environment.

    Parameters
    ----------
    env: gym.Env
        A gym environment, unwrappable to DiscreteEnv.
    with_absorbing_state: bool
        Use an absorbing state which is reached whenever a step returns done.

    Returns
    -------
    np.ndarray
        An array containing transition probabilities of shape (n_states, n_actions, n_states)
        or (n_states + 1, n_actions, n_states + 1) if an absorbing state is used.
    """
    env = unwrap_env(env, DiscreteEnv)

    n_states = env.observation_space.n
    if with_absorbing_state:
        # adding one state in the end of the table as absorbing state:
        n_states += 1
    n_actions = env.action_space.n

    table = np.zeros([n_states, n_actions, n_states])

    # Iterate over "from" states:
    for state, transitions_given_state in env.P.items():

        # Iterate over actions:
        for action, transitions in transitions_given_state.items():

            # Iterate over "to" states:
            for probability, next_state, _, done in transitions:
                table[state, action, next_state] += probability
                if done and with_absorbing_state is True:
                    # map next_state to absorbing state:
                    table[next_state, :, :] = 0.0
                    table[next_state, :, -1] = 1.0

    if with_absorbing_state:
        # absorbing state that is reached whenever done == True
        # only reaches itself for each action
        table[-1, :, -1] = 1.0

    return table


def get_reward_matrix(env: gym.Env, with_absorbing_state: bool = True):
    """Get reward array from discrete environment.

    If env is wrapped in a RewardWrapper, use the wrapper's rewards instead of
    the original ones.

    Parameters
    ----------
    env: gym.Env
        A gym environment, unwrappable to DiscreteEnv.
    with_absorbing_state: bool
        Use an absorbing state which is reached whenever a step returns done.

    Returns
    -------
    np.ndarray
        A numpy array of shape (n_states, n_actions) or (n_states + 1, n_actions) if
        with_absorbing_state == True. Contains values of R(s, a)
    """

    n_states = env.observation_space.n
    if with_absorbing_state:
        n_states += 1
    n_actions = env.action_space.n

    # unwrap the discrete env from which transitions can be extracted:
    discrete_env = unwrap_env(env, DiscreteEnv)
    # by default this discrete env's variable P will be used to extract rewards
    correct_transitions = copy(discrete_env.P)

    if with_absorbing_state:
        # change transitions in a way that terminal states map to the absorbing state
        for (state, transitions_for_state) in correct_transitions.items():
            for (action, transitions_for_state_action
                 ) in transitions_for_state.items():
                if len(transitions_for_state_action) == 1 \
                        and transitions_for_state_action[0][1] == state \
                        and transitions_for_state_action[0][3] is True:
                    # state maps to itself, but should map to absorbing state
                    rewired_outcome = (1.0, n_states - 1, 0, True)
                    correct_transitions[state][action] = [rewired_outcome]

    # however, if there is a reward wrapper, we need to use the
    # wrapped reward function:
    if is_unwrappable_to(env, RewardWrapper):
        # get the reward function:
        reward_wrapper = unwrap_env(env, RewardWrapper)
        reward_function = reward_wrapper.reward_function
        # re-calculate transitions based on reward function:
        transitions_based_on_reward_function = {}
        for (state, transitions_for_state) in correct_transitions.items():
            transitions_based_on_reward_function[state] = {}
            for (action, transitions_for_state_action
                 ) in transitions_for_state.items():
                outcomes = []
                for old_outcome in transitions_for_state_action:
                    outcome = list(copy(old_outcome))
                    next_state = outcome[1]
                    if with_absorbing_state and next_state == n_states - 1:
                        # hard coded: 0 reward when going to absorbing state
                        # (since the absorbing state is added artificially and
                        # not part of the wrapper's reward function)
                        outcome[2] = 0.0
                    else:
                        # update the reward part of the outcome
                        # reward function needs domain batch as input
                        assert reward_function.action_in_domain is False
                        assert reward_function.next_state_in_domain is False
                        reward_input = reward_wrapper.get_reward_input_for(
                            None, None, next_state)
                        outcome[2] = reward_function.reward(
                            reward_input).item()
                    outcomes.append(tuple(outcome))
                transitions_based_on_reward_function[state][action] = outcomes
        correct_transitions = transitions_based_on_reward_function

    rewards = np.zeros([n_states, n_actions])

    # Iterate over "from" states:
    for state, transitions_given_state in correct_transitions.items():

        # Iterate over actions:
        for action, transitions in transitions_given_state.items():

            # Iterate over "to" states:
            for probability, _, reward, _ in transitions:
                rewards[state, action] += reward * probability
    if with_absorbing_state:
        rewards[-1, :] = 0.0
    return rewards



"""Module containing true reward functions for different environments."""

import functools
import gym
import numpy as np

from irl_benchmark.irl.reward.reward_function import TabularRewardFunction

_true_reward_functions = {}


def make_true_reward(key: str):
    """Make a true reward function."""
    return _true_reward_functions[key]()


def _register_reward_function(key):
    def decorator(f):
        @functools.wraps(f)
        def new_f():
            return f(gym.make(key))

        new_f.__doc__ = "Creates true reward function for {}".format(key)
        _true_reward_functions[key] = new_f
        return new_f

    return decorator


@_register_reward_function('FrozenLake-v0')
def frozen_lake(env):
    parameters = np.zeros(16)
    parameters[-1] = 1.0
    return TabularRewardFunction(env, parameters)


@_register_reward_function('FrozenLake8x8-v0')
def frozen_lake_8_8(env):
    parameters = np.zeros(64)
    parameters[-1] = 1.0
    return TabularRewardFunction(env, parameters)



"""Module containing reward functions to be used for IRL."""
from abc import ABC, abstractmethod
from copy import copy
from typing import NamedTuple, Union

import gym
from gym.spaces.discrete import Discrete as DiscreteSpace
import numpy as np

from irl_benchmark.irl.feature.feature_wrapper import FeatureWrapper
import irl_benchmark.utils as utils


# Define some custom named tuples for easier handling
# of the reward function domains:
class State(NamedTuple):
    """A tuple containing an array of states."""
    state: np.ndarray


class StateAction(NamedTuple):
    """A tuple containing an array of states and an array of actions."""
    state: np.ndarray
    action: np.ndarray


class StateActionState(NamedTuple):
    """A tuple containing an array each for states, actions, and next states."""
    state: np.ndarray
    action: np.ndarray
    next_state: np.ndarray


class BaseRewardFunction(ABC):
    """The base class for all reward functions."""

    def __init__(self,
                 env: gym.Env,
                 parameters: Union[None, str, np.ndarray] = None,
                 action_in_domain: bool = False,
                 next_state_in_domain: bool = False):
        """ The abstract base class for reward functions

        Parameters
        ----------
        env: gym.Env
            A gym environment for which the reward function is defined.
        parameters: Union[None, str, np.ndarray]
            A numpy ndarray containing the parameters. If value is 'random',
            initializes with random parameters (mean 0, standard deviation 1).
        action_in_domain: bool
            Indicates whether actions are in the domain, i.e. R(s, a) or R(s, a, s')
        next_state_in_domain: bool
            Indicates whether next states are in the domain, i.e. R(s, a, s')
        """
        self.env = env
        self.action_in_domain = action_in_domain
        if next_state_in_domain:
            assert action_in_domain
        self.next_state_in_domain = next_state_in_domain
        self.parameters = parameters

    def domain(self) -> Union[State, StateAction, StateActionState]:
        """Return the entire domain of the reward function.

        Returns
        -------
        Union[State, StateAction, StateActionState]
            The domain of the reward function.
        """
        # domain always contains states:
        states = np.arange(self.env.observation_space.n)
        if self.action_in_domain:
            # if domain contains actions: extend domain
            states = np.repeat(states, self.env.action_space.n)
            actions = np.arange(self.env.action_space.n)
            actions = np.tile(actions, self.env.observation_space.n)
            if self.next_state_in_domain:
                # if domain contains next states: extend domain
                states = np.repeat(states, self.env.observation_space.n)
                actions = np.repeat(actions, self.env.observation_space.n)
                next_states = np.arange(self.env.observation_space.n)
                next_states = np.tile(
                    next_states,
                    self.env.observation_space.n * self.env.action_space.n)
                # return the adequate namedtuple:
                return StateActionState(states, actions, next_states)
            return StateAction(states, actions)
        return State(states)

    def domain_sample(self, batch_size: int
                      ) -> Union[State, StateAction, StateActionState]:
        """Return a batch sampled from the domain of size batch_size.

        Parameters
        ----------
        batch_size: int
            The size of the batch.

        Returns
        -------
        Union[State, StateAction, StateActionState]
            A batch sampled from the domain of the reward function.
        """
        raise NotImplementedError()

    @abstractmethod
    def reward(self, domain_batch: Union[State, StateAction, StateActionState]
               ) -> np.ndarray:
        """Return rewards for a domain batch.

        Parameters
        ----------
        domain_batch: Union[State, StateAction, StateActionState, np.ndarray]
            A batch of the domain (can in principle also be the entire domain).
            See :func:`domain` and :func:`domain_sample`.

        Returns
        -------
        np.ndarray
            Rewards for the given domain batch.
            The rewards are of shape (batch_size,)
        """
        raise NotImplementedError()

    def update_parameters(self, parameters: np.ndarray):
        """Update the parameters of the reward function.

        Parameters
        ----------
        parameters: np.ndarray
            The new parameters to be used in the reward function.
        """
        self.parameters = parameters


class TabularRewardFunction(BaseRewardFunction):
    """Rewards for each possible input are stored in a table.

    Only suitable for relatively small environments.
    The self.parameters in this case are the reward table's values.
    """

    def __init__(self,
                 env: gym.Env,
                 parameters: Union[None, str, np.ndarray] = None,
                 action_in_domain: bool = False,
                 next_state_in_domain: bool = False):
        """

        Parameters
        ----------
        env: gym.Env
            A gym environment for which the reward function is defined.
        parameters: Union[None, str, np.ndarray]
            A numpy ndarray containing the values for all elements in the reward table.
            If value is 'random', initializes with random parameters (mean 0, standard deviation 1).
            The size of parameters must correspond to the size of the domain
            (one table value for each possible input)
        action_in_domain: bool
            Indicates whether actions are in the domain, i.e. R(s, a) or R(s, a, s')
        next_state_in_domain: bool
            Indicates whether next states are in the domain, i.e. R(s, a, s')
        """
        super(TabularRewardFunction, self).__init__(
            env, parameters, action_in_domain, next_state_in_domain)

        # this reward function is only implemented for
        # discrete state and action spaces
        assert isinstance(env.observation_space, DiscreteSpace)
        assert isinstance(env.action_space, DiscreteSpace)
        # calculate number of elements in domain:
        self.domain_size = self.env.observation_space.n
        if self.action_in_domain:
            self.domain_size *= self.env.action_space.n
        if self.next_state_in_domain:
            self.domain_size *= self.env.observation_space.n

        if parameters is 'random':
            self.parameters = np.random.standard_normal(size=self.domain_size)
        else:
            self.parameters = np.array(parameters)
        assert len(self.parameters) == self.domain_size

    def domain_to_index(
            self, domain_batch: Union[State, StateAction, StateActionState]
    ) -> np.ndarray:
        """Convert a domain batch into  an numpy ndarray of reward table indices.

        Parameters
        ----------
        domain_batch: Union[State, StateAction, StateActionState]
            A domain batch. Can also be the entire domain.

        Returns
        -------
        np.ndarray
            A numpy array of corresponding reward table indices.
        """
        index = copy(domain_batch.state)
        if self.action_in_domain:
            index *= self.env.action_space.n
            index += domain_batch.action
            if self.next_state_in_domain:
                index *= self.env.observation_space.n
                index += domain_batch.next_state
        return index

    def reward(self, domain_batch: Union[State, StateAction, StateActionState]
               ) -> np.ndarray:
        """Return the corresponding rewards of a domain_batch.

        Parameters
        ----------
        domain_batch: Union[State, StateAction, StateActionState]
            A domain batch. Can also be the entire domain.

        Returns
        -------
        np.ndarray
            The rewards for a domain batch.
        """
        indices = self.domain_to_index(domain_batch)
        return self.parameters[indices]


class FeatureBasedRewardFunction(BaseRewardFunction):
    """A reward function which is linear in some provided features.

    The self.parameters are the coefficients that are multiplied with
    the features to get the reward (standard inner product).
    """

    def __init__(self,
                 env: gym.Env,
                 parameters: Union[None, str, np.ndarray] = None):
        """

        Parameters
        ----------
        env: gym.Env
            A gym environment. The environment has to be wrapped in a FeatureWrapper.
        parameters: Union[None, str, np.ndarray]
            The parameters of the reward function. One parameter for each feature.
            If value is 'random', initializes with random parameters (mean 0, standard deviation 1).
        """

        assert utils.wrapper.is_unwrappable_to(env, FeatureWrapper)
        super(FeatureBasedRewardFunction, self).__init__(env, parameters)

        if parameters is 'random':
            parameters_shape = utils.wrapper.unwrap_env(
                self.env, FeatureWrapper).feature_shape()
            self.parameters = np.random.standard_normal(parameters_shape)

    def reward_from_features(self, feature_batch: np.ndarray) -> np.ndarray:
        """Return corresponding rewards for a domain batch.

        Parameters
        ----------
        feature_batch: np.ndarray
            An array of features

        Returns
        -------
        np.ndarray
            The rewards for given features.
        """
        reward = np.dot(
            self.parameters.reshape(1, -1),
            feature_batch.reshape(len(self.parameters), -1)).reshape((-1))
        return reward

    def reward(self, domain_batch: Union[State, StateAction, StateActionState]
               ) -> np.ndarray:
        """Return rewards for a domain batch.
        Converts to features internally to calculate the reward.

        Parameters
        ----------
        domain_batch: Union[State, StateAction, StateActionState, np.ndarray]
            A batch of the domain (can in principle also be the entire domain).
            See :func:`domain` and :func:`domain_sample`.

        Returns
        -------
        np.ndarray
            Rewards for the given domain batch.
            The rewards are of shape (batch_size,)
        """
        features = self._domain_to_features(domain_batch)
        return self.reward_from_features(features)

    def _domain_to_features(
            self, domain_batch: Union[State, StateAction, StateActionState]
    ) -> np.ndarray:
        """Convert elements of the domain to features.

        Parameters
        ----------
        domain_batch: Union[State ,StateAction, StateActionState]
            A batch of elements from the reward function's domain.

        Returns
        -------
        np.ndarray
            A numpy array containing features.

        """
        feature_wrapper = utils.wrapper.unwrap_env(self.env, FeatureWrapper)
        feature_batch = []
        if isinstance(domain_batch.state, list):
            for i in range(len(domain_batch.state)):
                state = domain_batch.state[i]
                if self.action_in_domain:
                    action = domain_batch.action[i]
                else:
                    action = None
                if self.next_state_in_domain:
                    next_state = domain_batch.next_state[i]
                else:
                    next_state = None
                if not self.action_in_domain and not self.next_state_in_domain:
                    next_state = state
                    state = None
                features = feature_wrapper.features(state, action, next_state)
                feature_batch.append(features)
        else:
            state = domain_batch.state
            if self.action_in_domain:
                action = domain_batch.action
            else:
                action = None
            if self.next_state_in_domain:
                next_state = domain_batch.next_state
            else:
                next_state = None
            if not self.action_in_domain and not self.next_state_in_domain:
                next_state = state
                state = None
            features = feature_wrapper.features(state, action, next_state)
            feature_batch.append(features)
        return np.array(feature_batch)



"""Module for average trajectory return metric."""

from irl_benchmark.metrics.base_metric import BaseMetric
from irl_benchmark.utils.rl import true_reward_per_traj


class AverageTrajectoryReturn(BaseMetric):
    """Average sum of true rewards per trajectory."""

    def __init__(self, metric_input: dict):
        assert 'env' in metric_input.keys()
        super(AverageTrajectoryReturn, self).__init__(metric_input)

    def evaluate(self, evaluation_input: dict = None) -> float:
        """Evaluate the metric given some input and return result.

        Parameters
        ----------
        evaluation_input: dict
            The evaluation input. Requires 'irl_agent' field.

        Returns
        -------
        float
            The average sum of true rewards per trajectory.
        """
        assert 'irl_agent' in evaluation_input.keys()
        irl_trajs = self.generate_traj_if_not_exists(evaluation_input)
        return true_reward_per_traj(irl_trajs)



"""Module defining a run of an IRL algorithm experiment"""
from typing import Callable, Dict, List

import gym

from irl_benchmark.irl.algorithms.base_algorithm import BaseIRLAlgorithm
from irl_benchmark.irl.collect import load_stored_trajs
from irl_benchmark.irl.feature.feature_wrapper import make as feature_make
from irl_benchmark.irl.reward.reward_function import FeatureBasedRewardFunction, \
    TabularRewardFunction
from irl_benchmark.irl.reward.reward_wrapper import RewardWrapper
from irl_benchmark.irl.reward import truth
from irl_benchmark.metrics.base_metric import BaseMetric


class Run:
    """This class corresponds to a single run of an IRL algorithm experiment."""

    def __init__(self, env_id: str, expert_trajs_path: str,
                 irl_alg_factory: Callable[[gym.Env, List[Dict[str, list]]],
                                           BaseIRLAlgorithm],
                 metrics: List[BaseMetric], rl_config: dict, irl_config: dict,
                 run_config: dict):
        """

        Parameters
        ----------
        env_id: str
            The environment id of a gym environment. This is the id passed to gym.make().
        expert_trajs_path: str
            A path to the folder where expert trajectories are stored. The file with
            expert trajectories must be expert_trajs_path/trajs.data.
        irl_alg_factory: Callable[[gym.Env, List[Dict[str, list]]], BaseIRLAlgorithm]
            A factory function which takes a gym environment and expert trajetories and
            returns a subclass of BaseIRLAlgorithm.
        metrics: List[BaseMetric]
            The metrics to be evaluated after running the IRL algorithm.
        run_config: dict
            A dictionary containing the configuration of the run. Required fields are:
            'reward_function': subclass of BaseRewardFunction, e.g. FeatureBasedRewardFunction.
            'no_expert_trajs': int, number of expert trajectories to be used.
            'no_irl_iterations': int, number of iterations the IRL algorithm is run for.
            'no_rl_episodes_per_irl_iteration': int, how many episodes the RL agent
            is allowed to run each iteration.
            'no_irl_episodes_per_irl_iteration': int, how many episodes can be sampled
            for the IRL algorithm each iteration.
        """
        # create and wrap environment according to specified reward function
        if run_config['reward_function'] is FeatureBasedRewardFunction:
            # feature based reward functions need to be wrapped in a FeatureWrapper:
            self.env = feature_make(env_id)
        elif run_config['reward_function'] is TabularRewardFunction:
            self.env = gym.make(env_id)

        # wrap environment in a reward wrapper to prevent leaking of true reward
        reward_function = run_config['reward_function'](
            self.env, parameters='random')
        self.env = RewardWrapper(self.env, reward_function)

        # load expert trajs:
        self.expert_trajs = load_stored_trajs(expert_trajs_path)
        # use only specified number of expert trajs
        assert len(self.expert_trajs) >= run_config['no_expert_trajs']
        self.expert_trajs = self.expert_trajs[:run_config['no_expert_trajs']]

        self.irl_alg_factory = irl_alg_factory

        # Metrics are only passed as classes and need to be instantiated
        instantiated_metrics = []
        # collect all information relevant for certain metric __init__s:
        metric_input = {
            'env': self.env,
            'expert_trajs': self.expert_trajs,
            'true_reward': truth.make_true_reward(env_id),
        }
        # instantiate metrics:
        for metric in metrics:
            instantiated_metrics.append(metric(metric_input))
        self.metrics = instantiated_metrics

        self.rl_config = rl_config
        self.irl_config = irl_config

        self.run_config = run_config

    def start(self):
        """Start the run."""
        irl_alg = self.irl_alg_factory(self.env, self.expert_trajs,
                                       self.metrics, self.rl_config,
                                       self.irl_config)
        irl_alg.train(self.run_config['no_irl_iterations'],
                      self.run_config['no_rl_episodes_per_irl_iteration'],
                      self.run_config['no_irl_episodes_per_irl_iteration'])



"""Module for the base class of all metrics."""

from abc import ABC, abstractmethod

from irl_benchmark.irl.collect import collect_trajs


class BaseMetric(ABC):
    """The base class of all metrics."""

    def __init__(self, metric_input: dict = None):
        """

        Parameters
        ----------
        metric_input: dict
            This dictionary contains relevant data to initialize metrics.
            The dictionary is filled in each :class:`irl_benchmark.experiment.run.Run`.
            Available fields are 'env', 'expert_trajs', and 'true_reward'.
        """
        assert 'env' in metric_input.keys()
        assert 'expert_trajs' in metric_input.keys()
        assert 'true_reward' in metric_input.keys()
        self.metric_input = metric_input
        self.env = metric_input['env']

    @abstractmethod
    def evaluate(self, evaluation_input: dict = None) -> float:
        """Evaluate the metric given some input and return result.

        Parameters
        ----------
        evaluation_input: dict

        Returns
        -------
        float
            The metric for the given input
        """
        raise NotImplementedError()

    def generate_traj_if_not_exists(self, evaluation_input: dict):
        """ Generate trajectories and store them in evaluation input.
        If evaluation input already contains trajectories, do nothing.

        Parameters
        ----------
        evaluation_input

        Returns
        -------
        dict
            evaluation_input
        """
        assert 'irl_agent' in evaluation_input.keys()
        if not 'irl_trajs' in evaluation_input:
            print('generating new trajs for metrics')
            evaluation_input['irl_trajs'] = collect_trajs(
                self.env, evaluation_input['irl_agent'], 100)
        else:
            print('reuse generated trajs for metric')
        return evaluation_input['irl_trajs']



"""Module for Maximum Causal Entropy IRL."""
from typing import Callable, Dict, List

import gym
from gym.envs.toy_text.discrete import DiscreteEnv
import numpy as np

from irl_benchmark.config import IRL_CONFIG_DOMAINS
from irl_benchmark.irl.algorithms.base_algorithm import BaseIRLAlgorithm
from irl_benchmark.irl.feature.feature_wrapper import FeatureWrapper
from irl_benchmark.irl.reward.reward_function import FeatureBasedRewardFunction
from irl_benchmark.metrics.base_metric import BaseMetric
from irl_benchmark.rl.algorithms.base_algorithm import BaseRLAlgorithm
from irl_benchmark.utils.wrapper import get_transition_matrix, is_unwrappable_to, unwrap_env


class MaxCausalEntIRL(BaseIRLAlgorithm):
    """Maximum Entropy IRL (Ziebart et al., 2008).

    Not to be confused with Maximum Entropy Deep IRL (Wulfmeier et al., 2016)
    or Maximum Causal Entropy IRL (Ziebart et al., 2010).
    """
    def __init__(self, env: gym.Env, expert_trajs: List[Dict[str, list]],
                 rl_alg_factory: Callable[[gym.Env], BaseRLAlgorithm],
                 metrics: List[BaseMetric], config: dict):

        super(MaxCausalEntIRL, self).__init__(env, expert_trajs,
                                              rl_alg_factory, metrics, config)

        assert is_unwrappable_to(env, DiscreteEnv)
        assert is_unwrappable_to(env, FeatureWrapper)

        # get transition matrix (with absorbing state)
        self.transition_matrix = get_transition_matrix(self.env)
        self.n_states, self.n_actions, _ = self.transition_matrix.shape

        # get map of features for all states:
        feature_wrapper = unwrap_env(env, FeatureWrapper)
        self.feat_map = feature_wrapper.feature_array()

    def sa_visitations(self):
        """
        Given a list of trajectories in an MDP, computes the state-action
        visitation counts and the probability of a trajectory starting in state s.

            Arrays of shape (n_states, n_actions) and (n_states).
        """

        s0_count = np.zeros(self.n_states)
        sa_visit_count = np.zeros((self.n_states, self.n_actions))

        for traj in self.expert_trajs:
            # traj['states'][0] is the state of the first timestep of the trajectory.
            s0_count[traj['states'][0]] += 1

            for timestep in range(len(traj['actions'])):
                state = traj['states'][timestep]
                action = traj['actions'][timestep]

                sa_visit_count[state, action] += 1

        # Count into probability
        P0 = s0_count / len(self.expert_trajs)

        return sa_visit_count, P0

    def occupancy_measure(self, policy, P0, t_max=None, threshold=1e-6):
        """
        Computes occupancy measure of a MDP under a given time-constrained policy
        -- the expected discounted number of times that policy π visits state s in
        a given number of timesteps, as in Algorithm 9.3 of Ziebart's thesis:
        http://www.cs.cmu.edu/~bziebart/publications/thesis-bziebart.pdf.
        """

        if P0 is None:
            P0 = np.ones(self.n_states) / self.n_states
        d_prev = np.zeros_like(P0)

        t = 0

        diff = float("inf")

        while diff > threshold:

            d = np.copy(P0)

            for state in range(self.n_states):
                for action in range(self.n_actions):
                    # for all next_state reachable:

                    for next_state in range(self.n_states):
                        # probabilty of reaching next_state by taking action
                        prob = self.transition_matrix[state, action,
                                                      next_state]
                        d[next_state] += self.config['gamma'] * d_prev[
                            state] * policy[state, action] * prob

            diff = np.amax(abs(d_prev - d))  # maxima of the flattened array
            d_prev = np.copy(d)

            if t_max is not None:
                t += 1
                if t == t_max:
                    break
        return d_prev

    def train(self, no_irl_iterations: int,
              no_rl_episodes_per_irl_iteration: int,
              no_irl_episodes_per_irl_iteration: int):
        """

        """

        sa_visit_count, P0 = self.sa_visitations()

        # mean_s_visit_count = np.sum(sa_visit_count, 1) / len(self.expert_trajs)

        # calculate feature expectations
        expert_feature_count = self.feature_count(self.expert_trajs, gamma=1.0)

        # mean_feature_count = np.dot(self.feat_map.T, expert_feature_count )

        # initialize the parameters
        # theta = np.random.rand(self.feat_map.shape[1])

        reward_function = FeatureBasedRewardFunction(self.env, 'random')
        theta = reward_function.parameters

        agent = self.rl_alg_factory(self.env)

        irl_iteration_counter = 0

        while irl_iteration_counter < no_irl_iterations:
            irl_iteration_counter += 1

            if self.config['verbose']:
                print('IRL ITERATION ' + str(irl_iteration_counter))

            reward_function_estimate = FeatureBasedRewardFunction(
                self.env, theta)
            self.env.update_reward_function(reward_function_estimate)

            # compute policy
            agent.train(no_rl_episodes_per_irl_iteration)

            policy = agent.policy_array()
            state_values = agent.state_values
            q_values = agent.q_values

            # Log-Likelihood
            # l = np.sum(sa_visit_count * (q_values - state_values.T))  # check: broadcasting works as intended or not

            # occupancy measure
            d = self.occupancy_measure(policy=policy, P0=P0)[:-1]

            # log-likeilihood gradient
            grad = -(expert_feature_count - np.dot(self.feat_map.T, d))

            # graduate descent
            theta -= self.config['lr'] * grad

            evaluation_input = {
                'irl_agent': agent,
                'irl_reward': reward_function
            }
            self.evaluate_metrics(evaluation_input)

        return theta


IRL_CONFIG_DOMAINS[MaxCausalEntIRL] = {
    'gamma': {
        'type': float,
        'min': 0.0,
        'max': 1.0,
        'default': 0.9,
    },
    'verbose': {
        'type': bool,
        'default': True
    },
    'lr': {
        'type': float,
        'default': 0.02,
        'min': 0.000001,
        'max': 50
    }
}



from setuptools import setup

setup(
    name='irl-benchmark',
    version='0.1.0',
    packages=['irl_benchmark', 'irl_benchmark.rl', 'irl_benchmark.rl.algorithms', 'irl_benchmark.irl',
              'irl_benchmark.irl.reward', 'irl_benchmark.irl.collect', 'irl_benchmark.irl.feature',
              'irl_benchmark.irl.algorithms', 'irl_benchmark.utils', 'irl_benchmark.config', 'irl_benchmark.metrics',
              'irl_benchmark.experiment'],
    url='https://github.com/JohannesHeidecke/irl-benchmark',
    license='LICENSE',
    author='Anton Osika, Adria Garriga-Alonso, Johannes Heidecke, Max Daniel, Sayan Sarkar',
    author_email='jheidecke@gmail.com',
    description='Benchmark framework for inverse reinforcement learning algorithms.',
    install_requires=[
        'gym>=0.10.5',
        'cvxpy>=1.0.9',
        'numpy>=1.15.0',
        'typing>=3.6.4',
        'setuptools>=40.0.0',
        'comet_ml>=1.0.31',
        'msgpack_python>=0.5.6',
        'msgpack_numpy>=0.4.4.1',
        'torch>=0.4.0',
        'tqdm>=4.28.1',
    ],
    test_suite='nose.collector',
    tests_require=['nose'],
    include_package_data=True
)



"""Utils module containing general helper functions."""

from typing import Callable, List, Union

import numpy as np
import torch


def to_one_hot(hot_vals: Union[int, List[int], np.ndarray],
               max_val: int,
               zeros_function: Callable = np.zeros
               ) -> Union[np.ndarray, torch.tensor]:
    """ Convert an integer or a list of integers to a one-hot array.

    Parameters
    ----------
    hot_vals: Union[int, List[int], np.ndarray]
        A single integer, or a list / vector of integers, corresponding to the
        hot values which will equal one in the returned array.
    max_val: int
        The maximum possible value in hot_values. All elements in hot_vals have
        to be smaller than max_val (since we start counting at 0).
    zeros_function: Callable
        Controls which function is used to create the array. It should
        be either `numpy.zeros` or `torch.zeros`.

    Returns
    -------
    Union[np.ndarray, torch.tensor]
        Either a numpy array or torch tensor with the one-hot encoded values.
        Type of returned data structure depends on the passed zeros_function.
        The default is numpy array.
        The returned data structure will be of shape (1, max_value) if hot_vals
        is a single integer, and (len(hot_vals), max_value) otherwise.
    """
    assert np.max(hot_vals) < max_val
    assert np.min(hot_vals) >= 0
    try:
        n_rows = len(hot_vals)
        res = zeros_function((n_rows, max_val))
        res[np.arange(n_rows), hot_vals] = 1.
    except TypeError:
        res = zeros_function((max_val, ))
        res[hot_vals] = 1.
    return res



"""Module for the reward wrapper, hiding the original reward function and using a different one."""
from typing import Union, Tuple

import gym
import numpy as np

from irl_benchmark.irl.reward.reward_function import State
from irl_benchmark.irl.reward.reward_function import StateAction
from irl_benchmark.irl.reward.reward_function import StateActionState
from irl_benchmark.irl.reward.reward_function import BaseRewardFunction, \
    FeatureBasedRewardFunction, TabularRewardFunction


class RewardWrapper(gym.Wrapper):
    """Use a given reward function instead of the true reward provided by the environment"""

    def __init__(self, env: gym.Env, reward_function: BaseRewardFunction):
        """

        Parameters
        ----------
        env: gym.Env
            A gym environment (potentially already wrapped).
        reward_function: BaseRewardFunction
            The reward function to be used instead of the original reward.
        """
        super(RewardWrapper, self).__init__(env)
        self.reward_function = reward_function
        self.current_state = None

    def reset(self, **kwargs):
        """Call base class reset method and return initial state."""
        # pylint: disable=E0202
        self.current_state = self.env.reset()
        return self.current_state

    def step(self, action: Union[np.ndarray, int, float]
             ) -> Tuple[Union[np.ndarray, float, int], float, bool, dict]:
        """Call base class step method but replace reward with reward output by reward function.

        Parameters
        ----------
        action: Union[np.ndarray, int, float]
            An action, suitable for wrapped environment.

        Returns
        -------
        Tuple[Union[np.ndarray, float, int], float, bool, dict]
            Tuple with values for (state, reward, done, info).
            Normal return values of any gym step function. The reward value is replaced with
            the value output by the reward function given to the wrapper. The true reward is
            added as a field to the dictionary in the last element of the tuple
            with key 'true_reward'.
        """
        # pylint: disable=E0202

        # execute action:
        next_state, reward, terminated, info = self.env.step(action)

        # persist true reward in information:
        info['true_reward'] = reward

        # generate input for reward function:
        if isinstance(self.reward_function, FeatureBasedRewardFunction):
            # reward function can be called based on features
            reward = self.reward_function.reward_from_features(
                info['features']).item()
        elif isinstance(self.reward_function, TabularRewardFunction):
            rew_input = self.get_reward_input_for(self.current_state, action,
                                                  next_state)
            reward = self.reward_function.reward(rew_input).item()
        else:
            raise NotImplementedError()

        # remember which state we are in:
        self.current_state = next_state

        return next_state, reward, terminated, info

    def update_reward_function(self, reward_function):
        """Update the used reward function.

        Useful as IRL algorithms compute a new reward function
        in each iteration."""
        self.reward_function = reward_function

    def get_reward_input_for(self, state: Union[np.ndarray, int, float],
                             action: Union[np.ndarray, int, float],
                             next_state: Union[np.ndarray, int, float]
                             ) -> Union[State, StateAction, StateActionState]:
        """

        Parameters
        ----------
        state: Union[np.ndarray, int, float
        action: Union[np.ndarray, int, float]
        next_state: Union[np.ndarray, int, float]

        Returns
        -------
        Union[State, StateAction, StateActionState]
            The input converted to an adequate namedtuple.
        """
        if self.reward_function.action_in_domain:
            if self.reward_function.next_state_in_domain:
                return StateActionState(state, action, next_state)
            else:
                return StateAction(state, action)
        else:
            if not self.reward_function.action_in_domain and not self.reward_function.next_state_in_domain:
                state = next_state
            return State(state)



"""This module is used to collect trajectories for inverse reinforcement learning"""

import os
from typing import Dict, List, Union

import gym
import msgpack
import msgpack_numpy as m
from tqdm import tqdm

from irl_benchmark.rl.algorithms.base_algorithm import BaseRLAlgorithm

# set up MessagePack to be numpy compatible:
m.patch()


def collect_trajs(env: gym.Env,
                  agent: BaseRLAlgorithm,
                  no_trajectories: int,
                  max_steps_per_episode: Union[int, None] = None,
                  store_to: Union[str, None] = None,
                  verbose: bool = False) -> List[Dict[str, list]]:
    """ Collect and return trajectories of an agent in an environment.

    Parameters
    ----------
    env: gym.Env
        A gym environment
    agent: BaseRLAlgorithm
        An RL algorithm / agent
    no_trajectories: int
        Number of trajectories to be collected
    max_steps_per_episode: int
        Maximum number of steps allowed to take in each episode. Optional.
        If not set, the environment's default is used.
    store_to: Union[str, None]
        If not None: a path of where trajectories should be persisted.
    verbose: bool
        Whether to use a tqdm progress bar while collecting.

    Returns
    -------
    List[Dict[str, list]]
        A list of trajectories. Each trajectory is a dictionary with keys
        ['states', 'actions', 'rewards', 'true_rewards', 'features']. The
        values of each dictionary are lists. The list of states contains
        both starting and final state and is one element longer than the
        list of actions. The lists of true_rewards and features can be
        empty if the environment was not wrapped in a RewardWrapper or
        FeatureWrapper respectively.
    """

    # list of trajectories to be returned:
    trajectories = []

    for _ in (tqdm(range(no_trajectories))
              if verbose else range(no_trajectories)):
        # start new episode by resetting environment:
        state = env.reset()
        done = False

        # trajectories contain starting state.
        # this makes len(states) to be one larger than len(actions)
        states = [state]
        # initialize other lists:
        actions = []
        rewards = []
        true_rewards = []
        features = []

        # if no max steps are specified,
        # check if there is a default provided by the environment:
        if env.spec.max_episode_steps is not None:
            assert isinstance(env.spec.max_episode_steps, int)
            if max_steps_per_episode is None:
                max_steps_per_episode = env.spec.max_episode_steps
            elif env.spec.max_episode_steps < max_steps_per_episode:
                print(
                    'WARNING: running episodes for longer than the default of '
                    + str(env.spec.max_episode_steps) +
                    ' for this environment.')

        step_counter = 0
        while not done \
            and (max_steps_per_episode is None or step_counter <
                 max_steps_per_episode):
            step_counter += 1
            # Agent picks an action:
            action = agent.pick_action(state)
            # Action is executed by environment:
            next_state, reward, done, info = env.step(action)
            # Append obtained information to respective lists:
            states.append(next_state)
            actions.append(action)
            rewards.append(reward)
            if 'true_reward' in info.keys():
                true_rewards.append(info['true_reward'])
            if 'features' in info.keys():
                features.append(info['features'])
            # Update which state we are in:
            state = next_state

        # Construct trajectory dictionary:
        trajectory = {
            'states': states,
            'actions': actions,
            'rewards': rewards,
            'true_rewards': true_rewards,
            'features': features
        }
        # Append trajectory to list of trajectories:
        trajectories.append(trajectory)

    # If requested, store trajetories to file:
    if store_to is not None:
        store_trajs(trajectories, store_to)

    return trajectories


def store_trajs(trajs, store_to):
    """Store trajectories to store_to/trajs.data."""
    if not os.path.exists(store_to):
        os.makedirs(store_to)
    file_path = os.path.join(store_to, 'trajs.data')
    with open(file_path, 'wb') as file:
        msgpack.pack(trajs, file)


def load_stored_trajs(load_from):
    """Return trajectories storead at load_from/trajs.data."""
    file_path = os.path.join(load_from, 'trajs.data')
    with open(file_path, 'rb') as file:
        trajs = msgpack.unpack(file)
    # convert byte keys back to string keys:
    trajs = [{
        key.decode('utf-8') if isinstance(key, bytes) else key: traj[key]
        for key in traj.keys()
    } for traj in trajs]
    return trajs



"""Module for value iteration RL algorithm."""
from typing import Union

import gym
import numpy as np

from irl_benchmark.config import RL_CONFIG_DOMAINS
from irl_benchmark.rl.algorithms.base_algorithm import BaseRLAlgorithm
from irl_benchmark.utils.wrapper import is_unwrappable_to, \
    get_transition_matrix, get_reward_matrix


class ValueIteration(BaseRLAlgorithm):
    """Value iteration algorithm.

    Solves an MDP given exact knowledge of transition dynamics and rewards.
    Currently only implemented for DiscreteEnv environments.
    """

    def __init__(self, env: gym.Env, config: Union[None, dict] = None):
        """

        Parameters
        ----------
        env: gym.Env
            A DiscreteEnv environment
        config: dict
            Configuration of hyperparameters.
        """
        assert is_unwrappable_to(env, gym.envs.toy_text.discrete.DiscreteEnv)
        super(ValueIteration, self).__init__(env, config)
        self.no_states = env.observation_space.n + 1  # + 1 for absorbing state
        self.no_actions = env.action_space.n
        self.transitions = get_transition_matrix(env)
        # will be filled in beginning of training:
        self.rewards = None
        # will be filled during training:
        self.state_values = None
        self.q_values = None
        # whenever self._policy is None, it will be re-calculated
        # based on current self.q_values when calling policy().
        self._policy = None

    def train(self, no_episodes: int):
        """ Train the agent

        Parameters
        ----------
        no_episodes: int
            Not used in this algorithm (since it assumes known transition dynamics)
        """
        assert is_unwrappable_to(self.env,
                                 gym.envs.toy_text.discrete.DiscreteEnv)
        # extract reward function from env (using wrapped reward function if available):
        self.rewards = get_reward_matrix(self.env)

        # initialize state values:
        state_values = np.zeros([self.no_states])

        while True:  # stops when state values converge
            # remember old values for error computation
            old_state_values = state_values.copy()
            # calculate Q-values:
            q_values = self.rewards + \
                       self.config['gamma'] * self.transitions.dot(state_values)
            # calculate state values either with maximum or mellow maximum:
            if self.config['temperature'] is None:
                # using default maximum operator:
                state_values = self._argmax_state_values(q_values)
            else:
                # using softmax:
                state_values = self._softmax_state_values(q_values)

            # stopping condition:
            # check if state values converged (almost no change since last iteration:
            if np.allclose(
                    state_values, old_state_values,
                    atol=self.config['epsilon']):
                break

        # persist learned state values and Q-values:
        self.state_values = state_values
        self.q_values = q_values
        # flag to tell other methods that policy needs to be updated based on new values:
        self._policy = None

    def pick_action(self, state: int) -> int:
        """ Pick an action given a state.

        The way of picking is either uniformly random between all best options (argmax)
        or according to mellowmax distribution, if 'temperature' is not None in config.
        See :meth:`.policy`.

        Parameters
        ----------
        state: int
            An integer corresponding to a state of a DiscreteEnv.

        Returns
        -------
        int
            An action for a DiscreteEnv.
        """
        return np.random.choice(self.no_actions, p=self.policy(state))

    def policy(self, state: int) -> np.ndarray:
        """ Return the probabilities of picking all possible actions given a state.

        The probabilities are either uniformly random between all best options (argmax)
        or according to mellowmax distribution, if 'temperature' is not None in config.

        Parameters
        ----------
        state: int
            An integer corresponding to a state of a DiscreteEnv.

        Returns
        -------
        np.ndarray
            Action probabilities given the state.

        """
        assert self.q_values is not None, "Call train() before calling this method."
        assert np.isscalar(state)
        assert isinstance(state, (int, np.int64))

        return self.policy_array()[state, :]

    def policy_array(self):
        """Return action probabilities for all states as a numpy array.

        Returns
        -------
        np.ndarray
            Array containing probabilities for actions given a state.
            Shape: (n_states, n_actions)
        """
        self._update_policy_if_necessary()
        return self._policy

    def _update_policy_if_necessary(self):
        if self._policy is None:
            if self.config['temperature'] is None:
                self._policy = self._argmax_policy(self.q_values)
            else:
                self._policy = self._softmax_policy(self.q_values)

    def _argmax_policy(self, q_values: np.ndarray) -> np.ndarray:
        """ Calculate an argmax policy.

        Only picks actions with maximal Q-value given a state. If several actions
        are maxima, they are equally likely to be picked.

        Parameters
        ----------
        q_values:
            Q-values for all state-action pairs. Shape (n_states, n_actions)
        Returns
        -------
        np.ndarray
            Probabilities for actions given a state. Shape (n_states, n_actions)
        """
        # Find best actions:
        best_actions = np.isclose(q_values,
                                  np.max(q_values, axis=1).reshape((-1, 1)))
        # Initialize probabilities to be zero
        policy = np.zeros((self.no_states, self.no_actions))
        # Assign probability max to all best actions:
        policy[best_actions] = 1
        # Normalize values so their sum is 1. for each state:
        policy /= np.sum(policy, axis=1, keepdims=True)
        return policy

    def _softmax_policy(self, q_values):
        assert self.config['temperature'] is not None
        temperature = self.config['temperature']
        # for numerical stability (avoiding under- or overflow of exponent),
        # re-scale exponent without changing results of softmax,
        # using softmax(x) = softmax(x + c) for any constant c
        q_max = q_values.max(axis=1, keepdims=True)
        q_scaled = (q_values - q_max) / temperature
        # calculate softmax policy:
        policy = np.exp(q_scaled)
        # normalize values so their sum is 1. for each state:
        policy /= np.sum(policy, axis=1, keepdims=True)
        return policy

    def _mellowmax_policy(self, q_values):
        raise NotImplementedError()

    def _argmax_state_values(self, q_values):
        assert self.config['temperature'] is None
        return np.max(q_values, axis=1)

    def _softmax_state_values(self, q_values):
        assert self.config['temperature'] is not None
        # obtain probabilities of picking each (s, a):
        softmax_policy = self._softmax_policy(q_values)
        # multiply q_values by probability of picking them
        # then sum over actions to get state values:
        softmax_state_values = (softmax_policy * q_values).sum(axis=1)
        return softmax_state_values

    def _mellowmax_state_values(self, q_values):
        raise NotImplementedError()


RL_CONFIG_DOMAINS[ValueIteration] = {
    'gamma': {
        'type': float,
        'min': 0.0,
        'max': 1.0,
        'default': 0.9,
    },
    'epsilon': {
        'type': float,
        'min': 0.0,
        'max': float('inf'),
        'default': 1e-6,
    },
    'temperature': {
        'type': float,
        'optional': True,  # allows value to be None
        'min': 1e-10,
        'max': float('inf'),
        'default': None
    }
}
